//
// Author: Emin Ilker Cetinbas (niw3_at_yahoo_d0t_com)
// Generalization: Andras Varga
// Copyright (C) 2005 Emin Ilker Cetinbas, Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//


#include "MuseeMobility.h"
#include "FWMath.h"


#define MK_UPDATE  100
#define MK_CHANGE_DIR  101

Define_Module(MuseumMobility);

/****************************************************************
 * Construction functions.
 */
MuseumMobility::MuseumMobility()
{
    endTurn = NULL;
    endUpdate = NULL;
}

MuseumMobility::~MuseumMobility()
{
    cancelAndDelete(endTurn);
    cancelAndDelete(endUpdate);
}

/**
 * Reads the pausePeriod and the velocity
 *
 * If the host is not stationary it calculates a random position and
 * schedules a timer to trigger the first movement
 */
void MuseumMobility::initialize(int stage)
{
    BasicMobility::initialize(stage);

    EV << "initializing MuseumMobility stage " << stage << endl;

    if (stage == 0)
    {
        pausePeriod = &par("pausePeriod");
        walkingPeriod = &par("walkingPeriod");
		upDatePositionInterval = par("upDatePositionInterval");
        changeAngleInterval = &par("changeAngleInterval");
        changeAngleBy = &par("changeAngleBy");
        speed = &par("speed");
		

        // initalize self messages
		endTurn = new cMessage("turn", MK_CHANGE_DIR);//message used to update the direction
        endUpdate = new cMessage("update", MK_UPDATE);

        // initial speed and angle
        currentSpeed = speed->doubleValue();
        currentAngle = uniform(0, 360);
        step.x = currentSpeed * cos(PI * currentAngle / 180) * upDatePositionInterval;
        step.y = currentSpeed * sin(PI * currentAngle / 180) * upDatePositionInterval;

		State = PAUSE;
		walkingStart = lastUpdate = nextUpdate = 0;

		xMin = par("xMin");
		xMax = par("xMax");
		yMin = par("yMin");
		yMax = par("yMax");

        scheduleAt(simTime() + pausePeriod->doubleValue(), endUpdate);
        scheduleAt(simTime() + changeAngleInterval->doubleValue(), endTurn);
    }
}


/**
 * The only self message possible is to indicate a new movement.
 */
void MuseumMobility::handleSelfMsg(cMessage * msg)
{
    switch (msg->kind())
    {
    case MK_UPDATE:
		if (State == WALKING)
		{
//			currentAngle += changeAngleBy->doubleValue();
			if ((simTime() - walkingStart) < walkingPeriod->doubleValue())
			{
				currentSpeed = speed->doubleValue();
				step.x = currentSpeed * cos(PI * currentAngle / 180) * upDatePositionInterval;
				step.y = currentSpeed * sin(PI * currentAngle / 180) * upDatePositionInterval;
				move();
				updatePosition();
				nextUpdate = upDatePositionInterval;
			}
			else 
			{
				nextUpdate = pausePeriod->doubleValue();
				State = PAUSE;
			}
		}
		else 
		{
			State = WALKING;
			walkingStart = simTime();
			nextUpdate = upDatePositionInterval;
		}
		lastUpdate = simTime();
        scheduleAt(simTime() + nextUpdate, msg);
        break;
    case MK_CHANGE_DIR:
        currentAngle += changeAngleBy->doubleValue();
        scheduleAt(simTime() + changeAngleInterval->doubleValue(), msg);
        break;
    default:
        opp_error("Unknown self message kind in MuseumMobility class");
        break;
    }
}

/**
 * Move the host if the destination is not reached yet. Otherwise
 * calculate a new random position
 */
void MuseumMobility::move()
{
    pos += step;

    if (pos.x > xMax)
	{
	    pos.x -= 2*(pos.x - xMax);
        currentAngle = 180 - currentAngle;
	}
	if (pos.x < xMin)
	{
	    pos.x += 2*(xMin - pos.x);
        currentAngle = 180 - currentAngle;
	}
	if (pos.y > yMax)
	{
	    pos.y -= 2*(pos.y - yMax);
        currentAngle = -currentAngle;
	}
	if (pos.y < yMin)
	{
	    pos.y += 2*(yMin - pos.y);
        currentAngle = -currentAngle;
	}

    // do something if we reach the wall
    Coord dummy;
    handleIfOutside(REFLECT, dummy, step, currentAngle);

    EV << " xpos= " << pos.x << " ypos=" << pos.y << " speed=" << currentSpeed << endl;
}