//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#ifndef IEEE_80211_MAC_LBMS_H
#define IEEE_80211_MAC_LBMS_H

// uncomment this if you do not want to log state machine transitions
#define FSM_DEBUG

#include <list>
#include <vector>
#include "Ieee80211Mac.h"
#include "WirelessMacBase.h"
#include "IPassiveQueue.h"
#include "Ieee80211Frame_m.h"
#include "Ieee80211Consts.h"
#include "NotificationBoard.h"
#include "RadioState.h"
#include "FSMA.h"

/**
 *
 * @ingroup macLayer
 */
class INET_API LbmsMacSTA : public Ieee80211Mac
{
  struct MulticastMACAddress
  {
      MACAddress address;
      bool isLeader;
  };

  typedef std::vector<MulticastMACAddress> MulticastMACAddressList;
  typedef std::vector<MulticastMACAddress>::iterator MulticastMACAddressListI;

  protected:
    /**
     * @name Ieee80211Mac state variables
     * Various state information checked and modified according to the state machine.
     */
    //@{
    /**
     * A list MAC group address whitch MAC is associated with 
     */
    MulticastMACAddressList multicastGroupList;
    //@}

  protected:
    /** @name Statistics */
    //@{
    long numReceivedMulticast;
    int leaderID;
    cOutVector *Leader;
    //@}

  public:
    /**
     * @name Construction functions
     */
    //@{
    virtual ~LbmsMacSTA();
    //@}
  protected:
    /**
     * @name Initialization functions
     */
    //@{
    /** @brief Initialization of the module and its variables */
    virtual int numInitStages() const {return 2;}
    virtual void initialize(int);
    //@}

    protected:
    /**
     * @name Message handing functions
     * @brief Functions called from other classes to notify about state changes and to handle messages.
     */
    //@{
    /** @brief Handle messages from upper layer */
    virtual void handleUpperMsg(cMessage *msg);

    /** @brief Handle all kinds of messages and notifications with the state machine */
    virtual void handleWithFSM(cMessage *msg);
    //@}

    protected:
    /**
     * @name Timer functions
     * @brief These functions have the side effect of starting the corresponding timers.
     */
    //@{
    void scheduleMulticastTimeoutPeriod(Ieee80211DataOrMgmtFrame *frame);
    //@}

  protected:
    /**
     * @name Frame transmission functions
     */
    //@{
    void sendMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend);
    //@}

  protected:
    /**
     * @name Frame builder functions
     */
    //@{
    Ieee80211DataOrMgmtFrame *buildMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend);
    //@}

  protected:
    /**
     * @name Utility functions
     */
    //@{
    /** @brief Returns true if message is a multicast message */
    bool isMulticast(Ieee80211Frame *msg); 

    /** @brief Returns true if message destination address is ours */
    bool isForUs(Ieee80211Frame *msg);

    /** @brief Returns true if the MAC is leader */
    bool isLeader(Ieee80211Frame *frame);

    /** @brief Set this MAC as Leader */
    void setLeader(Ieee80211Frame *frame);

    /** @brief Join a mutlicast group */
    void joinMulticastGroup(Ieee80211DataOrMgmtFrame *frame);

    /** @brief Leave a mutlicast group */
    void leaveMulticastGroup(Ieee80211DataOrMgmtFrame *frame);

    /** @brief Returns true if the MAC multicast adresse is a new one */
    bool isNewGroup(Ieee80211Frame *frame);

    /** @brief Process Incoming LBMS frame */
    void processIncomingLBMS(Ieee80211Frame *frameToSend);

    /** @brief Logs MultiCastGroup */
    void logMultiCast();

    //@}
};

#endif
