//
// Copyright (C) 2006 Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//


#include "Ieee80211MgmtAP.h"
#include "Ieee802Ctrl_m.h"
#include "EtherFrame_m.h"
#include "NotifierConsts.h"
#include "RadioState.h"


Define_Module(Ieee80211MgmtAP);


#define NOTIFY_ASSOCIATION			// to enable/disable notification mechanism between APs
#define BUFFER_FRAMES				// to enable/disable transmission of buffered frames at APs
//#define CENTRALIZED_BUFFERING
#define DISTRIBUTED_BUFFERING

#define MAX_NUM_FRAMES_TO_SEND  16	// maximum number of buffered frames to send at a time
#define TX_TIMER 0.05				// time period of sending buffered frames
#define TTL_TIMER  1

static std::ostream& operator<< (std::ostream& os, const Ieee80211MgmtAP::STAInfo& sta)
{
    os << "state:" << sta.status;
    return os;
}

void Ieee80211MgmtAP::initialize(int stage)
{
    Ieee80211MgmtAPBase::initialize(stage);

    if (stage==0)
    {
        // read params and init vars
        ssid = par("ssid").stringValue();
        beaconInterval = par("beaconInterval");
        numAuthSteps = par("numAuthSteps");
        if (numAuthSteps!=2 && numAuthSteps!=4)
            error("parameter 'numAuthSteps' (number of frames exchanged during authentication) must be 2 or 4, not %d", numAuthSteps);
        channelNumber = -1;  // value will arrive from physical layer in receiveChangeNotification()
        WATCH(ssid);
        WATCH(channelNumber);
        WATCH(beaconInterval);
        WATCH(numAuthSteps);
        WATCH_MAP(staList);

        //TBD fill in supportedRates

        // subscribe for notifications
        NotificationBoard *nb = NotificationBoardAccess().get();
        nb->subscribe(this, NF_RADIO_CHANNEL_CHANGED);

        // start beacon timer (randomize startup time)
        beaconTimer = new cMessage("beaconTimer");
        scheduleAt(uniform(0,beaconInterval), beaconTimer);
		txTimer = new cMessage("txTimer");
		ttlTimer = new cMessage("ttlTimer");
		sendFramesFlag = FALSE;
		buffer.listBufferLimit_ = par("bufferSize");
		bufferingEnabled = par("bufferingEnable");
		ttlValue = par("ttl");
		globalStoreCounter = 0;
    }
}

void Ieee80211MgmtAP::handleTimer(cMessage *msg)
{
    if (msg==beaconTimer)
    {
        sendBeacon();
        scheduleAt(simTime()+beaconInterval, beaconTimer);
    }
	else if (msg==txTimer)
	{// if the timer is for transmission of buffered frames
		// send stored frames to destination "tempAddress"
		std::cout << "********* global store counter for ap (" << myAddress << ") = " << globalStoreCounter << " *********\n";
		if (buffer.listSize_ > 0)
			SendStoreFramesToDest();
		if (buffer.listSize_ > 0)
		{
			scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames
		}
		else
		{
			cancelEvent(txTimer);
		}		
	}
	else if (msg==ttlTimer)
	{
		if (buffer.listSize_ > 0)
		{
			DecrementTTLValuesFromBuffer();
			scheduleAt(simTime()+TTL_TIMER, ttlTimer);
		}
		else
		{
			cancelEvent(ttlTimer);
		}
	}
    else
    {
        error("internal error: unrecognized timer '%s'", msg->name());
    }
}

void Ieee80211MgmtAP::handleUpperMessage(cMessage *msg)
{
#ifdef NOTIFY_ASSOCIATION
	/***********************************************************************************************/
	/* this part of the code is for the case when notification mechanism is enabled in between APs */
	int status;
	int numFramesSent;
	Ieee80211DataFrame frameNotif;

	MACAddress destAddress;
	MACAddress apAddress;

	int numStoredFrames;

#ifdef CENTRALIZED_BUFFERING
	MACAddress storageStation;
	storageStation.setAddress("F4-00-FF-0F-AA-20");
#endif

	// handle notification message came from Ethernet and get the destination address info, and number of frames already sent
	status = hNotif.CheckNotificationMsg(msg, &destAddress, &apAddress, &numFramesSent, &frameNotif, myAddress);
	STAList::iterator itAss;	// association list iterator to check the current status of a station

	// check the type of the current notification message received
	switch(status)
	{
	case ASS_NOTIF:
	case DEST_FOUND_NOTIF:
		// association notification: a node is associated with any other AP in the network
		// it may be possible that this AP still considers that node to be associated with itself
		// this may be because the node didn't send an explicit disassociation before leaving (we don't have timeout for association)
		// so when an AP receives an association notification from another AP and the node is still associated with it, we should disassociated the node from this AP (in the list)
		// find the destination address in the associated list at this AP
		itAss = staList.find(destAddress);	

//		std::cout << "low priority drops = " << buffer.lowPriorityLossCounter_ << endl;
//		std::cout << "high priority drops = " << buffer.highPriorityLossCounter_ << endl;

		// check if the station is found
		if (itAss != staList.end())
		{
			// the destination address has an entry in AP's associated list
			// check if the station status is still "associated"
			if (itAss->second.status == ASSOCIATED)
			{
				// change the station (node)'s status to "authenticated"
				itAss->second.status = AUTHENTICATED;
			}
		}
		return;		// we should not parse through handleUpperMsg any more in case of reception of notifications
		break;

	case DISASS_NOTIF:
		// disassociation notification: a node is disassociated from another AP in the network
		if (buffer.listSize_ == 0)
		{
			if (txTimer)
				cancelEvent(txTimer);
		}
		return;		// we should not parse through handleUpperMsg any more in case of reception of notifications
		break;

	case UPDATE_FRAMES_NOTIF:
		return;		// we should not parse through handleUpperMsg any more in case of reception of notifications
		break;
	
	case REMOVE_FRAMES_NOTIF:
		return;		// we should not parse through handleUpperMsg any more in case of reception of notifications
		break;

	case STORE_FRAMES:
#ifdef CENTRALIZED_BUFFERING
		if (storageStation == myAddress)
		{
#endif
			// buffer frames here
			if (bufferingEnabled)
			{
//				std::cout << "Treating Store Frames\n";
#ifdef DISTRIBUTED_BUFFERING
/*				std::list<MACAddress>::iterator destLocalListIterator;
				for (destLocalListIterator = destBufferLocalList.begin(); destLocalListIterator != destBufferLocalList.end(); destLocalListIterator++)
				{
					if (*destLocalListIterator == destAddress)
					{
*/				
				std::list<DestBufferInfo>::iterator destBufferInfoIterator;
				for (destBufferInfoIterator = hNotif.destBufferInfoList.begin(); destBufferInfoIterator != hNotif.destBufferInfoList.end(); destBufferInfoIterator++)
				{
					if (destBufferInfoIterator->destNodeAddress == destAddress && destBufferInfoIterator->respAPAddress == myAddress)
					{
#endif
						int priority = frameNotif.getAID();
						bufferedFrames_t frameForBuffer;
						frameForBuffer.frame = frameNotif;
						frameForBuffer.priority = priority;
						frameForBuffer.ttl = ttlValue;
						buffer.AddElementToList(frameForBuffer);
						globalStoreCounter++;
						if (ttlTimer)
						{
							
						}
						else
						{
							std::cout << "setting ttl timer\n";
							scheduleAt(simTime()+TTL_TIMER, ttlTimer);
						}
#ifdef DISTRIBUTED_BUFFERING
						break;
					}
				}
#endif
			}
#ifdef CENTRALIZED_BUFFERING
		}
#endif
		return;
		break;

	case FETCH_FRAMES:		
#ifdef CENTRALIZED_BUFFERING
		if (storageStation == myAddress)
		{
#endif
			// count the number of stored frames for the destination
			numStoredFrames = CountNumberOfStoredFrames(destAddress);			
			std::cout << "Number of Stored Frames for dest = " << destAddress << " are " << numStoredFrames << "\n";
			std::cout << "Total Number of Frames at " << myAddress << " = " << buffer.listSize_ << endl;

			// check if there are still some frames to be sent for the destination "tempAddress"
			if (numStoredFrames > 0)
			{
				if (txTimer)
					cancelEvent(txTimer);
				// send frames to the corresponding AP
				SendStoreFramesToDest();
				scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames				
			}
#ifdef CENTRALIZED_BUFFERING
		}
#endif
		return;
		break;

	default:
//		delete apNotifMsg;
		break;
	}
#endif

	Ieee80211DataFrame *frame;
	// must be an EtherFrame frame arriving from MACRelayUnit, that is,
    // bridged from another interface of the AP (probably Ethernet).
    EtherFrame *etherframe = check_and_cast<EtherFrame *>(msg);

	frame = (Ieee80211DataFrame*) etherframe->decapsulate();
	if (!frame)
		error("received empty EtherFrame from upper layer");

	frame->setToDS(false);
	frame->setFromDS(true);
	frame->setReceiverAddress(frame->getAddress3());
	frame->setAddress3(frame->getTransmitterAddress());
	frame->setTransmitterAddress(myAddress);

	// check we really have a STA with that dest address
    STAList::iterator it = staList.find(frame->getReceiverAddress());

    if (it==staList.end() || it->second.status!=ASSOCIATED)
    {
		if (!etherframe->getDest().isMulticast())
        {// The case when an AP receives a data frame from another AP for a destination which is not associated
//			we need to buffer this frame here
//			copy addresses from ethernet frame (transmitter addr will be set to our addr by MAC)
//			frame->setReceiverAddress(ethframe->getDest());
//			frame->setAddress3(ethframe->getSrc());

			EV << "Ieee80211MgmtAP::handleUpperMessage, Searching for destination info " << frame->getReceiverAddress() << " \n";
			
			if (frame->getReceiverAddress() == MACAddress::BROADCAST_ADDRESS)
			{
				EV << "It is possibly an ARP frame. Sending the packet to lower layer \n";
			}

			else 
			{
#ifdef NOTIFY_ASSOCIATION
/*				if (hNotif.CheckForDestInfo(frame->getReceiverAddress(), &apAddress) == DEST_INFO_ENTRY_NOT_FOUND)
				{
					std::cout << "Checking if statement" << endl;
					int priority = frame->getAID();
//					std::cout << "priority = " << priority << endl;
					bufferedFrames_t frameForBuffer;
					frameForBuffer.frame = *frame;
					frameForBuffer.priority = priority;

					if (bufferingEnabled)
						buffer.AddElementToList(frameForBuffer);

				}
*/
				delete frame;
				delete etherframe;
				return;
#endif
			}
        }
        else
             EV << "Sending Multicast message \n";
    }

	// convert Ethernet frame

	EV << "Ieee80211MgmtAP::handleUpperMessage, Transmitting data frame\n";
//	Ieee80211DataFrame *frame = convertFromEtherFrame(etherframe);
	delete etherframe;
    sendOrEnqueue(frame);
	
	// TODO: Send the notification over Ethernet to all APs telling the destination is found here
	// Destination Found Notification
}

void Ieee80211MgmtAP::handleCommand(int msgkind, cPolymorphic *ctrl)
{
    error("handleCommand(): no commands supported");
}

void Ieee80211MgmtAP::receiveChangeNotification(int category, cPolymorphic *details)
{
    Enter_Method_Silent();
    printNotificationBanner(category, details);

    if (category == NF_RADIO_CHANNEL_CHANGED)
    {
        EV << "updating channel number\n";
        channelNumber = check_and_cast<RadioState *>(details)->getChannelNumber();
    }
}

Ieee80211MgmtAP::STAInfo *Ieee80211MgmtAP::lookupSenderSTA(Ieee80211ManagementFrame *frame)
{
    STAList::iterator it = staList.find(frame->getTransmitterAddress());
    return it==staList.end() ? NULL : &(it->second);
}

void Ieee80211MgmtAP::sendManagementFrame(Ieee80211ManagementFrame *frame, const MACAddress& destAddr)
{
    frame->setFromDS(true);
    frame->setReceiverAddress(destAddr);
    frame->setAddress3(myAddress);
    sendOrEnqueue(frame);
}

void Ieee80211MgmtAP::sendBeacon()
{
    EV << "Sending beacon\n";
    Ieee80211BeaconFrame *frame = new Ieee80211BeaconFrame("Beacon");
    Ieee80211BeaconFrameBody& body = frame->getBody();
    body.setSSID(ssid.c_str());
    body.setSupportedRates(supportedRates);
    body.setBeaconInterval(beaconInterval);
    body.setChannelNumber(channelNumber);

    frame->setReceiverAddress(MACAddress::BROADCAST_ADDRESS);
    frame->setFromDS(true);

//	send(frame, "macOut"); 	
	sendOrEnqueue(frame);
}

void Ieee80211MgmtAP::handleDataFrame(Ieee80211DataFrame *frame)
{
    // check toDS bit
    if (!frame->getToDS())
    {
        // looks like this is not for us - discard
        EV << "frame->getToDS()" <<  frame->getToDS() <<endl;
        delete frame;
        return;
    }

    // look up destination address in our STA list
    EV << " AP frame->getAddress3() " << frame->getAddress3() <<endl;
    STAList::iterator it = staList.find(frame->getAddress3());
	MACAddress apAddress;

    if (it==staList.end())
    {
        // not our STA -- pass up frame to relayUnit for LAN bridging if we have one
        if (hasRelayUnit)
		{
			if (frame->getAddress3() == MACAddress::BROADCAST_ADDRESS)
			{
				Ieee80211DataFrame* frame1 = (Ieee80211DataFrame*) frame->dup();
				distributeReceivedDataFrame(frame1);
				
				EtherFrame *ethframe = new EthernetIIFrame(frame->name()); //TODO option to use EtherFrameWithSNAP instead
				ethframe->setDest(frame->getAddress3());
				ethframe->setSrc(frame->getTransmitterAddress());
				//XXX set ethertype
				ethframe->setKind(ETH_FRAME);//set kind to ETH_FRAME so that ETH MAC may work
				ethframe->encapsulate(frame);

				send(ethframe, "uppergateOut");
			}
			
			else
			{
#ifdef NOTIFY_ASSOCIATION
				if (hNotif.CheckForDestInfo(frame->getAddress3(), &apAddress) == DEST_INFO_ENTRY_FOUND)
				{
				    // create a matching ethernet frame
					EtherFrame *ethframe = new EthernetIIFrame(frame->name()); //TODO option to use EtherFrameWithSNAP instead
					ethframe->setDest(apAddress);
					ethframe->setSrc(frame->getTransmitterAddress());
					//XXX set ethertype
					ethframe->setKind(ETH_FRAME);//set kind to ETH_FRAME so that ETH MAC may work					
					ethframe->encapsulate(frame);

					send(ethframe, "uppergateOut");
				}
				else
				{
					// create a matching ethernet frame
					EtherFrame* payload;
//					std::cout << "Sending STORE_FRAMES Notification\n";
					// prepare notification frame to be sent to all APs over Ethernet telling about the current status of buffer for this station
					MACAddress destAddress = frame->getAddress3();
//					payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);

#ifdef CENTRALIZED_BUFFERING
					payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);
					MACAddress apStAddress;
					apStAddress.setAddress("F4-00-FF-0F-AA-20");
					payload->setDest(apStAddress);
					send(payload, "uppergateOut");
#endif
#ifdef DISTRIBUTED_BUFFERING
					MACAddress tempAPAddress;
					if (hNotif.CheckForDestBufferInfo(destAddress, &tempAPAddress) == DEST_BUFFER_INFO_ENTRY_FOUND)
					{
						if (tempAPAddress != myAddress)
						{
							payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);
							payload->setDest(tempAPAddress);
							send(payload, "uppergateOut");
						}
						else
						{
							if (bufferingEnabled)
							{
								int priority = frame->getAID();
								bufferedFrames_t frameForBuffer;
								frameForBuffer.frame = *frame;
								frameForBuffer.priority = priority;
								frameForBuffer.ttl = ttlValue;
								buffer.AddElementToList(frameForBuffer);
								std::cout << "buffering data for " << destAddress << endl;
								globalStoreCounter++;
								if (ttlTimer)
								{
									
								}
								else
								{
									std::cout << "setting ttl timer\n";
									scheduleAt(simTime()+TTL_TIMER, ttlTimer);
								}
								delete frame;
							}
						}
					}
					else
					{
						std::cout << "Check 7 8 9\n";
					}
#endif					
				}
#endif
			}
		}
        else
            distributeReceivedDataFrame(frame);
            //delete frame;
    }
    else
    {
        // dest address is our STA, but is it already associated?
        if (it->second.status == ASSOCIATED)
		{
			std::cout << "Forwarding the Frame Locally at sim time " << simTime() << " for destination " << frame->getAddress3() << "\n";
            distributeReceivedDataFrame(frame); // send it out to the destination STA
		}
        else
		{
//			std::cout << "deleting the frame\n";
			if (hNotif.CheckForDestInfo(frame->getAddress3(), &apAddress) == DEST_INFO_ENTRY_FOUND)
			{
				// create a matching ethernet frame
				EtherFrame *ethframe = new EthernetIIFrame(frame->name()); //TODO option to use EtherFrameWithSNAP instead
				ethframe->setDest(apAddress);
				ethframe->setSrc(frame->getTransmitterAddress());
				//XXX set ethertype
				ethframe->setKind(ETH_FRAME);//set kind to ETH_FRAME so that ETH MAC may work					
				ethframe->encapsulate(frame);

				send(ethframe, "uppergateOut");
			}
			else
			{
				EV << "Ieee80211MgmtAP::handleDataFrame, Buffering the frame before broadcasting\n";
				// create a matching ethernet frame
				EtherFrame* payload;
//				std::cout << "Sending STORE_FRAMES Notification\n";
				// prepare notification frame to be sent to all APs over Ethernet telling about the current status of buffer for this station
				MACAddress destAddress = frame->getAddress3();
//				payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);
#ifdef CENTRALIZED_BUFFERING
				payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);
				MACAddress apStAddress;
				apStAddress.setAddress("F4-00-FF-0F-AA-20");
				payload->setDest(apStAddress);
				send(payload, "uppergateOut");
#endif
#ifdef DISTRIBUTED_BUFFERING
				MACAddress tempAPAddress;
				if (hNotif.CheckForDestBufferInfo(destAddress, &tempAPAddress) == DEST_BUFFER_INFO_ENTRY_FOUND)
				{
					if (tempAPAddress != myAddress)
					{
						payload = hNotif.SendNotificationToAPs(STORE_FRAMES, frame->getAddress3(), myAddress, 0, frame);
						payload->setDest(tempAPAddress);						
						send(payload, "uppergateOut");
					}
					else
					{
						if (bufferingEnabled)
						{
							int priority = frame->getAID();
							bufferedFrames_t frameForBuffer;
							frameForBuffer.frame = *frame;
							frameForBuffer.priority = priority;
							frameForBuffer.ttl = ttlValue;
							buffer.AddElementToList(frameForBuffer);
							std::cout << "buffering data for " << destAddress << endl;
							globalStoreCounter++;
							if (ttlTimer)
							{
								
							}
							else
							{
								std::cout << "setting ttl timer\n";
								scheduleAt(simTime()+TTL_TIMER, ttlTimer);
							}
							delete frame;
						}
					}					
				}
				else
				{
					std::cout << "Check 7 8 9\n";
				}
#endif
			}
//            delete frame;
		}
    }
}

void Ieee80211MgmtAP::handleAuthenticationFrame(Ieee80211AuthenticationFrame *frame)
{
    int frameAuthSeq = frame->getBody().getSequenceNumber();
    EV << "Processing Authentication frame, seqNum=" << frameAuthSeq << "\n";

    // create STA entry if needed
    STAInfo *sta = lookupSenderSTA(frame);
    if (!sta)
    {
        EV << "Station not present in the list. Creating an entry\n";
		MACAddress staAddress = frame->getTransmitterAddress();
        sta = &staList[staAddress]; // this implicitly creates a new entry
        sta->address = staAddress;
        sta->status = NOT_AUTHENTICATED;
        sta->authSeqExpected = 1;
    }

#ifdef NOTIFY_ASSOCIATION
	if (sta->status == ASSOCIATED)
	{

		if (hasRelayUnit)
		{
			EtherFrame* payload;
			std::cout << "Sending Disassociation Notification\n";			

			if (buffer.listSize_ == 0)
			{
				if (txTimer)
					cancelEvent(txTimer);
			}
			// TODO: Here buffer.listSize_ should be replaced by total number of frames for current destination
			
			int numStoredFrames = CountNumberOfStoredFrames(sta->address);
			payload = hNotif.SendNotificationToAPs(DISASS_NOTIF, sta->address, myAddress, numStoredFrames, NULL);
		
//			std::cout << "Current List Size = " << buffer.listSize_ << " No of Stored Frames = " << numStoredFrames << endl;
			send(payload, "uppergateOut");
		}
	}
#endif

	EV << "handleAuthenticationFrame::Station Status is " << sta->status << "\n";

    // check authentication sequence number is OK
    if (frameAuthSeq != sta->authSeqExpected)
    {
        // wrong sequence number: send error and return
        EV << "Wrong sequence number, " << sta->authSeqExpected << " expected\n";
        Ieee80211AuthenticationFrame *resp = new Ieee80211AuthenticationFrame("Auth-ERROR");
        resp->getBody().setStatusCode(SC_AUTH_OUT_OF_SEQ);
        sendManagementFrame(resp, frame->getTransmitterAddress());
        delete frame;
        sta->authSeqExpected = 1; // go back to start square
        return;
    }

    // station is authenticated if it made it through the required number of steps
    bool isLast = (frameAuthSeq+1 == numAuthSteps);

    // send OK response (we don't model the cryptography part, just assume
    // successful authentication every time)
    EV << "Sending Authentication frame, seqNum=" << (frameAuthSeq+1) << "\n";
    Ieee80211AuthenticationFrame *resp = new Ieee80211AuthenticationFrame(isLast ? "Auth-OK" : "Auth");
    resp->getBody().setSequenceNumber(frameAuthSeq+1);
    resp->getBody().setStatusCode(SC_SUCCESSFUL);
    resp->getBody().setIsLast(isLast);
    // XXX frame length could be increased to account for challenge text length etc.
    sendManagementFrame(resp, frame->getTransmitterAddress());

    delete frame;

    // update status
    if (isLast)
    {
		// to handle the case where a station is disassociated but without notifying AP about it
		// (may be due to loss of beacons)
		// so here the AP changes its status directly from ASSOCIATED to AUTHENTICATED
		// so we notify here all other APs that the station is diassociated
		
		sta->status = AUTHENTICATED; // XXX only when ACK of this frame arrives
        EV << "STA authenticated\n";
//		std::cout << "STA " << sta->address << " authenticated\n";
/*
		if (hasRelayUnit)
		{
			EtherFrame* payload;
			payload = hNotif.SendNotificationToAPs(ASS_NOTIF, sta->address, myAddress);
	
			send(payload, "uppergateOut");
		}
*/
    }
    else
    {
        sta->authSeqExpected += 2;
        EV << "Expecting Authentication frame " << sta->authSeqExpected << "\n";
    }
}

void Ieee80211MgmtAP::handleDeauthenticationFrame(Ieee80211DeauthenticationFrame *frame)
{
    EV << "Processing Deauthentication frame\n";

    STAInfo *sta = lookupSenderSTA(frame);
    delete frame;

    if (sta)
    {
        // mark STA as not authenticated; alternatively, it could also be removed from staList
        sta->status = NOT_AUTHENTICATED;
        sta->authSeqExpected = 1;
    }
}

void Ieee80211MgmtAP::handleAssociationRequestFrame(Ieee80211AssociationRequestFrame *frame)
{
    EV << "Processing AssociationRequest frame\n";

    // "11.3.2 AP association procedures"
    STAInfo *sta = lookupSenderSTA(frame);
    if (!sta || sta->status==NOT_AUTHENTICATED)
    {
        EV << "Association Request arrived but station not authenticated\n";
		// STA not authenticated: send error and return
        Ieee80211DeauthenticationFrame *resp = new Ieee80211DeauthenticationFrame("Deauth");
        resp->getBody().setReasonCode(RC_NONAUTH_ASS_REQUEST);
        sendManagementFrame(resp, frame->getTransmitterAddress());
        delete frame;
        return;
    }

    delete frame;

	EV << "handleAssociationRequestFrame::Station Status is " << sta->status << "\n";

    // mark STA as associated
    sta->status = ASSOCIATED; // XXX this should only take place when MAC receives the ACK for the response

    // send OK response
    Ieee80211AssociationResponseFrame *resp = new Ieee80211AssociationResponseFrame("AssocResp-OK");
    Ieee80211AssociationResponseFrameBody& body = resp->getBody();
    body.setStatusCode(SC_SUCCESSFUL);
    body.setAid(0); //XXX
    body.setSupportedRates(supportedRates);
    sendManagementFrame(resp, sta->address);

#ifdef NOTIFY_ASSOCIATION
	if (hasRelayUnit)
	{
		EtherFrame* payload;
		payload = hNotif.SendNotificationToAPs(ASS_NOTIF, sta->address, myAddress, 0, 0);
		send(payload, "uppergateOut");

		hNotif.AddOrModifyDestInfo(sta->address, myAddress);

		payload = hNotif.SendNotificationToAPs(FETCH_FRAMES, sta->address, myAddress, 0, NULL);

#ifdef CENTRALIZED_BUFFERING
		MACAddress apStAddress;
		apStAddress.setAddress("F4-00-FF-0F-AA-20");
		payload->setDest(apStAddress);
		send(payload, "uppergateOut");
#endif

#ifdef DISTRIBUTED_BUFFERING
		payload->setDest(MACAddress::BROADCAST_ADDRESS);
		send(payload, "uppergateOut");

		int numStoredFrames = CountNumberOfStoredFrames(sta->address);

		std::cout << "handleAssociationRequestFrame(), Number of Stored Frames for dest " << sta->address << " is = " << numStoredFrames << endl;

		if (numStoredFrames > 0)
		{
			if (txTimer)
				cancelEvent(txTimer);
			SendStoreFramesToDest();
			scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames
		}

/*		MACAddress apAddress;
		if (hNotif.CheckForDestBufferInfo(sta->address, &apAddress) == DEST_BUFFER_INFO_ENTRY_FOUND)
		{
			if (apAddress == myAddress)
			{
				// count the number of stored frames for the destination
				int numStoredFrames = CountNumberOfStoredFrames(sta->address);
				std::cout << "Number of Stored Frames for dest = " << sta->address << " are " << numStoredFrames << "\n";
				std::cout << "Total Number of Frames = " << buffer.listSize_ << endl;

				// check if there are still some frames to be sent for the destination "tempAddress"
				if (txTimer)
					cancelEvent(txTimer);
				if (numStoredFrames > 0)
				{
					SendStoreFramesToDest();
					scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames					
				}
			}
			else
			{
				payload->setDest(apAddress);
				send(payload, "uppergateOut");
			}
		}

		else
		{
			int status = hNotif.CheckAndAddDestBufferAP(sta->address, myAddress);
			destBufferLocalList.push_back(sta->address);
		}
*/
#endif
	}
#endif
}

void Ieee80211MgmtAP::handleAssociationResponseFrame(Ieee80211AssociationResponseFrame *frame)
{
    dropManagementFrame(frame);
}

void Ieee80211MgmtAP::handleReassociationRequestFrame(Ieee80211ReassociationRequestFrame *frame)
{
    EV << "Processing ReassociationRequest frame\n";

    // "11.3.4 AP reassociation procedures" -- almost the same as AssociationRequest processing
    STAInfo *sta = lookupSenderSTA(frame);
    if (!sta || sta->status==NOT_AUTHENTICATED)
    {
        EV << "ReAssociation Request arrived but station not authenticated\n";
		// STA not authenticated: send error and return
        Ieee80211DeauthenticationFrame *resp = new Ieee80211DeauthenticationFrame("Deauth");
        resp->getBody().setReasonCode(RC_NONAUTH_ASS_REQUEST);
        sendManagementFrame(resp, frame->getTransmitterAddress());
        delete frame;
        return;
    }

    delete frame;

	EV << "handleReassociationRequestFrame::Station Status is " << sta->status << "\n";

    // mark STA as associated
    sta->status = ASSOCIATED; // XXX this should only take place when MAC receives the ACK for the response

    // send OK response
    Ieee80211ReassociationResponseFrame *resp = new Ieee80211ReassociationResponseFrame("ReassocResp-OK");
    Ieee80211ReassociationResponseFrameBody& body = resp->getBody();
    body.setStatusCode(SC_SUCCESSFUL);
    body.setAid(0); //XXX
    body.setSupportedRates(supportedRates);
    sendManagementFrame(resp, sta->address);

#ifdef NOTIFY_ASSOCIATION
	if (hasRelayUnit)
	{
		EtherFrame* payload;
		payload = hNotif.SendNotificationToAPs(ASS_NOTIF, sta->address, myAddress, 0, 0);
		send(payload, "uppergateOut");

		hNotif.AddOrModifyDestInfo(sta->address, myAddress);

		payload = hNotif.SendNotificationToAPs(FETCH_FRAMES, sta->address, myAddress, 0, NULL);

#ifdef CENTRALIZED_BUFFERING
		MACAddress apStAddress;
		apStAddress.setAddress("F4-00-FF-0F-AA-20");
		payload->setDest(apStAddress);
		send(payload, "uppergateOut");
#endif

#ifdef DISTRIBUTED_BUFFERING
		payload->setDest(MACAddress::BROADCAST_ADDRESS);
		send(payload, "uppergateOut");

		int numStoredFrames = CountNumberOfStoredFrames(sta->address);

		if (numStoredFrames > 0)
		{
			if (txTimer)
				cancelEvent(txTimer);
			SendStoreFramesToDest();
			scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames
		}

/*		MACAddress apAddress;
		if (hNotif.CheckForDestBufferInfo(sta->address, &apAddress) == DEST_BUFFER_INFO_ENTRY_FOUND)
		{
			if (apAddress == myAddress)
			{
				// count the number of stored frames for the destination
				int numStoredFrames = CountNumberOfStoredFrames(sta->address);
				std::cout << "Number of Stored Frames for dest = " << sta->address << " are " << numStoredFrames << "\n";
				std::cout << "Total Number of Frames = " << buffer.listSize_ << endl;

				// check if there are still some frames to be sent for the destination "tempAddress"
				if (txTimer)
					cancelEvent(txTimer);
				if (numStoredFrames > 0)
				{
					SendStoreFramesToDest();
					scheduleAt(simTime()+TX_TIMER, txTimer);	// reschedule the timer for transmission of buffered frames					
				}
			}
			else
			{
				payload->setDest(apAddress);
				send(payload, "uppergateOut");
			}
		}

		else
		{
			int status = hNotif.CheckAndAddDestBufferAP(sta->address, myAddress);
			destBufferLocalList.push_back(sta->address);
		}
*/
#endif		
	}
#endif
}

void Ieee80211MgmtAP::handleReassociationResponseFrame(Ieee80211ReassociationResponseFrame *frame)
{
    dropManagementFrame(frame);
}

void Ieee80211MgmtAP::handleDisassociationFrame(Ieee80211DisassociationFrame *frame)
{
    STAInfo *sta = lookupSenderSTA(frame);
    delete frame;

	EV << "Received DisAssociation Request\n";
    if (sta)
    {
        sta->status = AUTHENTICATED;
    }

#ifdef NOTIFY_ASSOCIATION
	if (hasRelayUnit)
	{
		EtherFrame* payload;
		std::cout << "Sending Disassociation Notification for station " << sta->address << "\n";

		if (buffer.listSize_ == 0)
		{
			if (txTimer)
				cancelEvent(txTimer);
		}

		payload = hNotif.SendNotificationToAPs(DISASS_NOTIF, sta->address, myAddress, 0, NULL);
		send(payload, "uppergateOut");

		hNotif.RemoveDestInfo(sta->address);
#ifdef DISTRIBUTED_BUFFERING
		hNotif.CheckAndAddDestBufferAP(sta->address, myAddress);
#endif
	}
#endif
}

void Ieee80211MgmtAP::handleBeaconFrame(Ieee80211BeaconFrame *frame)
{
    dropManagementFrame(frame);
}

void Ieee80211MgmtAP::handleProbeRequestFrame(Ieee80211ProbeRequestFrame *frame)
{
    EV << "Processing ProbeRequest frame\n";

    if (strcmp(frame->getBody().getSSID(),"")!=0 && strcmp(frame->getBody().getSSID(), ssid.c_str())!=0)
    {
        EV << "SSID `" << frame->getBody().getSSID() << "' does not match, ignoring frame\n";
        dropManagementFrame(frame);
        return;
    }

    MACAddress staAddress = frame->getTransmitterAddress();
    delete frame;

    EV << "Sending ProbeResponse frame\n";
    Ieee80211ProbeResponseFrame *resp = new Ieee80211ProbeResponseFrame("ProbeResp");
    Ieee80211ProbeResponseFrameBody& body = resp->getBody();
    body.setSSID(ssid.c_str());
    body.setSupportedRates(supportedRates);
    body.setBeaconInterval(beaconInterval);
    body.setChannelNumber(channelNumber);
    sendManagementFrame(resp, staAddress);
}

void Ieee80211MgmtAP::handleProbeResponseFrame(Ieee80211ProbeResponseFrame *frame)
{
    dropManagementFrame(frame);
}


void Ieee80211MgmtAP::SendStoreFramesToDest()
{
	Ieee80211DataFrame *ptrThisFrame;
	int counter = 0;
	int i = 0;
	MACAddress destAddress;
	MACAddress tempDestAddress[20];
	int tempDestAddressSize=0;

#ifdef CENTRALIZED_BUFFERING
	std::list<struct DestInfo>::iterator destInfoIterator;		// link list iterator
	for (destInfoIterator = hNotif.destInfoList.begin(); destInfoIterator != hNotif.destInfoList.end(); destInfoIterator++)
	{
		destAddress = destInfoIterator->destNodeAddress;
		int count = CountNumberOfStoredFrames(destAddress);
//		std::cout << "dest address = " << destAddress << " ||| count = " << count << " ||| total = " << buffer.listSize_ << endl;
		if (count > 0)
		{
			tempDestAddress[i++] = destAddress;
			tempDestAddressSize++;
		}
	}

	std::cout << "dest list size = " << tempDestAddressSize << endl;

	for (int j=0, i=0; i<tempDestAddressSize; )
	{
		destAddress = tempDestAddress[j];
//		std::cout << "SendStoreFramesToDest(): dest address = " << destAddress << endl;
#endif

#ifdef DISTRIBUTED_BUFFERING
//	std::cout << "local list size = " << destBufferLocalList.size() << endl;

	bool flag = FALSE;
	std::list<DestInfo>::iterator destInfoIterator;
	MACAddress apAddress;
	for (destInfoIterator = hNotif.destInfoList.begin(), i=0; ; )
	{
		destAddress = destInfoIterator->destNodeAddress;
		apAddress = destInfoIterator->assAPAddress;
/*
	std::list<MACAddress>::iterator destBufferLocalListIterator;		// link list iterator
	for (destBufferLocalListIterator = destBufferLocalList.begin(), i=0; ; )
	{
		destAddress = *destBufferLocalListIterator;
*/
//		std::cout << "SendStoreFramesToDest(): dest address = " << destAddress << " my address = " << myAddress << endl;		
#endif

		int count = CountNumberOfStoredFrames(destAddress);
//		std::cout << "SendStoreFramesToDest(): Count for dest " << destAddress << " is " << count << endl;
		
		if (count > 0)
		{
			std::list<bufferedFrames_t>::iterator newIterator;		// link list iterator
			for (newIterator = buffer.ptrPktList_.begin(); newIterator != buffer.ptrPktList_.end(); )
			{
				thisFrame = (Ieee80211DataFrame) (newIterator->frame);

				if (newIterator->frame.getToDS())
				{
					if (destAddress == newIterator->frame.getAddress3())
					{
#ifdef CENTRALIZED_BUFFERING
						MACAddress apAddress;
						hNotif.CheckForDestBufferInfo(destAddress, &apAddress);
#endif
#ifdef DISTRIBUTED_BUFFERING
//						if (hNotif.CheckForDestInfo(destAddress, &apAddress) == DEST_INFO_ENTRY_FOUND)
//						{
							flag = TRUE;
//							std::cout << "dest address = " << destAddress << " ap address = " << apAddress << " buffer ap = " << apStAddress << " my address = " << myAddress << endl;
							if (apAddress != myAddress)
							{
#endif
								EtherFrame *ethframe = new EthernetIIFrame((&thisFrame)->name()); //TODO option to use EtherFrameWithSNAP instead
								ethframe->setDest(apAddress);
								ethframe->setSrc(thisFrame.getTransmitterAddress());
								//XXX set ethertype
								ethframe->setKind(ETH_FRAME);//set kind to ETH_FRAME so that ETH MAC may work					
								ethframe->encapsulate((Ieee80211DataFrame*)(&thisFrame)->dup());
								send(ethframe, "uppergateOut");
								newIterator = buffer.ptrPktList_.erase(newIterator);
								(buffer.listSize_)--;
								i++;
								break;
#ifdef DISTRIBUTED_BUFFERING
							}
							else
							{
								thisFrame.setToDS(false);
								thisFrame.setFromDS(true);

								ASSERT(!thisFrame.getAddress3().isUnspecified());
								
								thisFrame.setReceiverAddress(destAddress);
								thisFrame.setAddress3(thisFrame.getTransmitterAddress());
								thisFrame.setTransmitterAddress(myAddress);
								sendOrEnqueue((Ieee80211DataFrame*)(&thisFrame)->dup());
								newIterator = buffer.ptrPktList_.erase(newIterator);
								(buffer.listSize_)--;
								i++;
								break;
							}							
/*						}
						else
						{
							counter++;
							newIterator++;
						}
*/
#endif						
					}
					else
					{
//						EV << "Dest Address = " << destAddress << " doesn't match. Checking next entry\n";
						counter++;
						newIterator++;
					}
				}
				else
				{
					if (destAddress == newIterator->frame.getReceiverAddress())
					{
						std::cout << "SendStoreFramesToDest(): get receiver address =====> " << newIterator->frame.getReceiverAddress() << endl;
#ifdef DISTRIBUTED_BUFFERING
						flag = TRUE;
						if (apAddress != myAddress)
						{
#endif
							thisFrame.setTransmitterAddress(myAddress);

							EtherFrame *ethframe = new EthernetIIFrame((&thisFrame)->name()); //TODO option to use EtherFrameWithSNAP instead
							ethframe->setDest(MACAddress::BROADCAST_ADDRESS);
							ethframe->setSrc(thisFrame.getTransmitterAddress());
							//XXX set ethertype
							ethframe->setKind(ETH_FRAME);//set kind to ETH_FRAME so that ETH MAC may work					
							ethframe->encapsulate((Ieee80211DataFrame*)(&thisFrame)->dup());
							send(ethframe, "uppergateOut");

		//					sendOrEnqueue((Ieee80211DataFrame*)(&thisFrame)->dup());

							newIterator = buffer.ptrPktList_.erase(newIterator);
							(buffer.listSize_)--;
							i++;
							break;
#ifdef DISTRIBUTED_BUFFERING
						}
						else
						{
							sendOrEnqueue((Ieee80211DataFrame*)(&thisFrame)->dup());
							newIterator = buffer.ptrPktList_.erase(newIterator);
							(buffer.listSize_)--;
							i++;
							break;
						}
#endif
					}
					else
					{
	//					EV << "Dest Address = " << destAddress << " doesn't match. Checking next entry\n";
						counter++;
						newIterator++;
					}
				}
			}
		}

#ifdef DISTRIBUTED_BUFFERING
		destInfoIterator++;		
#endif
#ifdef CENTRALIZED_BUFFERING
		j++;
		if (i == (MAX_NUM_FRAMES_TO_SEND*tempDestAddressSize) || buffer.ptrPktList_.size() == 0)
#endif
#ifdef DISTRIBUTED_BUFFERING
		if (i == (MAX_NUM_FRAMES_TO_SEND) || buffer.ptrPktList_.size() == 0)
#endif
		{				
			break;
		}
#ifdef DISTRIBUTED_BUFFERING
		if (destInfoIterator == hNotif.destInfoList.end())
		{
			if (flag)
			{
				destInfoIterator = hNotif.destInfoList.begin();
				flag = FALSE;
			}
			else
			{
				break;
			}
		}

#endif
#ifdef CENTRALIZED_BUFFERING
		if (j == tempDestAddressSize)
		{
			j=0;
		}
#endif
	}
}	

void Ieee80211MgmtAP::RemoveFramesFromBufferForDest(MACAddress destAddress, int numFramesToRemove)
{
	Ieee80211DataFrame *ptrThisFrame;
	int counter = 0;
	int removeCounter = 0;
	
	std::list<bufferedFrames_t>::iterator newIterator;		// link list iterator

//	std::cout << "MAC Address = " << destAddress << endl;

	std::cout << "Removing " << numFramesToRemove << " Packets at " << myAddress << " List Size = " << buffer.listSize_ << "\n";
	for (newIterator = buffer.ptrPktList_.begin(); newIterator != buffer.ptrPktList_.end(); )
	{
		if (removeCounter == numFramesToRemove)
				break;

		if (newIterator->frame.getToDS())
		{
			if (destAddress == newIterator->frame.getAddress3())
			{
				newIterator = buffer.ptrPktList_.erase(newIterator);
				(buffer.listSize_)--;
				removeCounter++;
			}
			else
			{
				newIterator++;
//				EV << "Dest Address " << destAddress << " doesn't match. Checking next entry\n";
				counter++;
			}
		}
		else
		{
			if (destAddress == newIterator->frame.getReceiverAddress())
			{
				newIterator = buffer.ptrPktList_.erase(newIterator);
				(buffer.listSize_)--;
				removeCounter++;
			}
			else
			{
				newIterator++;
//				EV << "Dest Address " << destAddress << " doesn't match. Checking next entry\n";
				counter++;
			}
		}
	}
}

int Ieee80211MgmtAP::CountNumberOfStoredFrames(MACAddress destAddress)
{
	int counter = 0;
	
	std::list<bufferedFrames_t>::iterator newIterator;

	for (newIterator = buffer.ptrPktList_.begin(); newIterator != buffer.ptrPktList_.end(); newIterator++)
	{
		thisFrame = (Ieee80211DataFrame) (newIterator->frame);

		if (newIterator->frame.getToDS())
		{
			if (destAddress == newIterator->frame.getAddress3())
			{
				counter++;
			}
		}
		else
		{
			if (destAddress == newIterator->frame.getReceiverAddress())
			{
				counter++;
			}
		}
	}

	EV << "Number of Stored Frames for dest = " << destAddress << " are " << counter << "\n";
	return counter;
}

void Ieee80211MgmtAP::DecrementTTLValuesFromBuffer(void)
{
	std::cout << "DecrementTTLValuesFromBuffer(), Decrementing TTL\n";

	std::list<bufferedFrames_t>::iterator newIterator;
	for (newIterator = buffer.ptrPktList_.begin(); newIterator != buffer.ptrPktList_.end(); )
	{
		newIterator->ttl--;
		if (newIterator->ttl == 0)
		{
			std::cout << "DecrementTTLValuesFromBuffer(), TTL expired - Dropping Frame\n";
			newIterator = buffer.ptrPktList_.erase(newIterator);
		}
		else
		{
			newIterator++;
		}
	}
}

Ieee80211MgmtAP::~Ieee80211MgmtAP()
{
	if (txTimer)
	{
		cancelEvent(txTimer);
		delete txTimer;
	}
	if (ttlTimer)
	{
		cancelEvent(ttlTimer);
		delete ttlTimer;
	}
//	delete beaconTimer;
}
